VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "AWE_Sheet"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
'===================================================================================================
' Class Name: AWE_Sheet
' Overview: Handles up to 10K rows with sub-second performance, robust error handling, and industry-
'           standard practices that streamline your code, allowing you to focus on solutions, not syntax.
'
'     FREE: Download AWE_Sheet today at www.AutomateWithExcel.com and revolutionize Excel automation!
'
' Description: Automates Excel worksheet operations with performance, flexibility, and robust error handling.
'              Provides functionality for workbooks, worksheets, and tables (ListObjects), including:
'              - Column mapping
'              - Filtering and sorting
'              - Data manipulation and validation
'              Optimized for local, network, or SharePoint-based worksheets. Compatible with Microsoft
'              Excel 365 and earlier versions. AWE_Sheet eliminates manual errors and streamlines
'              repetitive tasks with centralized validation.
'
' Author: Mike Libby
'   With over 30 years of development expertise and GPT-4 collaboration, this tool delivers a
'   simplified approach to Excel automation for results-driven solutions.
'
' Version: 2024.11.27
' Date Created: November 2024
' Last Modified: 2024.11.27
'
' Change Log:
' - Initial release: Core workbook/worksheet functionality, column mapping, filtering, sorting,
'   range management, and error handling.
'
' Dependencies:
' - Microsoft Excel 365 (or compatible versions)
' - VBA runtime environment
'
' Primary Feature:
'   AWE_Sheet streamlines code and dynamically adapts to both ListObjects (tables) and ranges,
'   ensuring consistent functionality across formats.
'
' Key Features:
' - Sub-second performance for datasheets up to 10K rows.
' - Centralized error handling and validation for reliable operations.
' - Flexible column mapping and dynamic data retrieval.
' - Built-in methods for sorting, filtering, lookups, and range management.
' - Modular design for simple and complex automation tasks.
' - Automates repetitive Excel tasks effortlessly.
' - Reduces manual errors with seamless task and report automation.
' - Ideal for project managers, data analysts, and Excel power users.
' - Fully compatible with local, SharePoint, and network files.
' - Quickly filter, sort, and manipulate data without cumbersome VBA coding.
' - Scales seamlessly from small datasets to large projects with thousands of rows.
'
' License: Freeware (personal or commercial use with attribution)
'===================================================================================================

'===================================================================================================
' Quick Start Example: Filter rows where "Status" is "Completed" or "In Progress" and iterate visible rows.
'   On Error GoTo ErrorHandler
'   Dim aweSh As New AWE_Sheet
'
'   ' Initialize AWE_Sheet for the target worksheet
'   aweSh.Initialize vTarget:="Sheet or Tab Name", iHdrRowNb:=2
'
'   ' Apply filter to "Status" column
'   aweSh.FilterX "Status", Array("Completed", "In Progress")
'
'   ' Iterate through rows visible after applying the filter
'   Dim rowRng As Range
'   For Each rowRng In aweSh.DataBodyRangeX ' Iterate visible rows only
'       Debug.Print aweSh.CellsX(rowRng, "Employee Name").Value
'   Next rowRng
'
'   ' See the many alternative methods to retrieve data using FindRowNumbers and CellsX
'
'   Exit Sub
'
' ErrorHandler:
'   'ErrorHandler: Captures and raises detailed error information, including number, source, and description.
'   RaiseError Err.Number, "Your Function's Name", "Error applying filter: " & Err.Description
'   ' Example of detailed error information:
'   '   Err.Description = "Error applying filter: Column Name was not found - Status"
'   '   Err.Source = "Your Function's Name -> XFilter -> Columns"
'===================================================================================================

'***************************************************************************************************
'* Design Rules: Seamlessly handle both ListObjects (tables) and regular ranges.                   *
'*               Supports one ListObject or table per sheet.                                       *
'*               All ranges, counts, lastrow... routines return visible (xlCellTypeVisible) rows.  *
'*               Filtered or fragmented ranges are consolidated and ghost rows are excluded.       *
'*               Seamlessly handle contiguous and non-contiguous rows across all functions.        *
'*               Error handling provides call stack traceability.                                  *
'*               Streamline calls to advanced functionality for downstream systems.                *
'*                                                                                                 *
'*               FOCUS ON SOLUTIONS, NOT SYNTAX                                                    *
'***************************************************************************************************

Option Explicit

' Core Variables
Private m_ParentWorkbook As Workbook   ' Associated workbook
Private m_ParentWorksheet As Worksheet ' Associated worksheet
Private m_HdrRowNb As Long             ' Header row number
Private m_ListObject As ListObject     ' Associated table (if any)
Private m_ColumnHdrDict As Object      ' Dictionary for column name-to-number mapping

'###################################################################################################
' Error Handling - Codes and RaiseError Function
'###################################################################################################
'---------------------------------------------------------------------------------------------------
' Enum Name: AWE_SheetError
' Purpose: Error codes to handle common issues with workbooks, worksheets, and metadata.
'---------------------------------------------------------------------------------------------------
Public Enum AWE_SheetError
    ERR_WorkbookNotFound = 1001            ' Workbook not found.
    ERR_InvalidWorksheet = 1002            ' Worksheet missing or invalid.
    ERR_HeaderRowOutOfRange = 1003         ' Header row out of range.
    ERR_MetadataInitializationError = 1004 ' Column metadata mapping failed.
    ERR_ColumnNotFound = 1005              ' Column not found.
    ERR_SheetEmpty = 1006                  ' Worksheet is empty.
    ERR_InvalidTable = 1007                ' ListObject (table) invalid.
    ERR_ProtectedWorksheet = 1008          ' Worksheet protected, operation blocked.
    ERR_RangeValidationFailed = 1009       ' Range or array validation failed.
    ERR_SearchNotFound = 1010              ' Search data not found.
    ERR_InvalidParameter = 1011            ' Unsupported parameter type or invalid value passed.
End Enum

'---------------------------------------------------------------------------------------------------
' RaiseError - Handles all errors with a unified format and traceability.
' Params:  errorCode (AWE_SheetError) - The error code to raise.
'          functionName (String) - Name of the calling function for error context.
'          description (String) - Detailed error message to display.
'---------------------------------------------------------------------------------------------------
Public Sub RaiseError(errorCode As AWE_SheetError, functionName As String, Description As String)
    Dim fullSource As String
    fullSource = "AWE_Sheet::" & functionName & IIf(Err.Source <> vbNullString, "->" & Err.Source, "")
    Err.Raise errorCode, fullSource, Description
End Sub

'---------------------------------------------------------------------------------------------------
' CError - Returns the numeric value for the specified AWE_SheetError enumeration.
'---------------------------------------------------------------------------------------------------
Public Property Get CError(EnumName As AWE_SheetError) As Long
    CError = EnumName
End Property

'###################################################################################################
' Initialization - Functions that instantiate AWE_Sheet
'###################################################################################################
'---------------------------------------------------------------------------------------------------
' Initialize - Prepares the AWE_Sheet object for workbook/worksheet operations. Validates inputs and
'              initializes metadata for column mapping.
' Params:  SheetNameOrObj (Variant) - Worksheet, sheet name, or ListObject.
'          HeaderRowNumber (Optional, Default: 1) - Header row number.
'          WorkbookFileNameOrObj (Optional, Default: ThisWorkbook) - Workbook object/file name.
'          OpenReadOnly (Optional, Default: False) - Open workbook as read-only.
' Returns: AWE_Sheet - Initialized instance or raises error on failure.
'
' Example: aweSh.Initialize SheetNameOrObj:="Sheet1"                ' Initialize with a local sheet.
'          aweSh.Initialize SheetNameOrObj:="RemoteSheet", _        ' Initialize a remote sheet.
'              WorkbookFileNameOrObj:="C:\Files\RemoteWorkbook.xlsx", OpenReadOnly:=True
'          aweSh.Initialize SheetNameOrObj:=ActiveSheet.ListObjects(1)  ' Initialize with a ListObject.
'
' Notes:   - Reuses open workbook if available to avoid duplicate openings.
'          - Automatically detects and integrates ListObjects on the target sheet.
'---------------------------------------------------------------------------------------------------
Public Function Initialize(SheetNameOrObj As Variant, _
                           Optional HeaderRowNumber As Long = 1, _
                           Optional WorkbookFileNameOrObj As Variant = Nothing, _
                           Optional OpenReadOnly As Boolean = False) As AWE_Sheet
    On Error GoTo ErrorHandler

    ' Clear all references and reset variables
    If m_ColumnHdrDict Is Nothing Then
        Set m_ColumnHdrDict = CreateObject("Scripting.Dictionary")
        m_ColumnHdrDict.CompareMode = vbTextCompare
    End If
    m_ColumnHdrDict.RemoveAll
    Set m_ParentWorkbook = Nothing
    Set m_ParentWorksheet = Nothing
    Set m_ListObject = Nothing
    m_HdrRowNb = 0

    ' Step 1: Set Workbook
    Select Case TypeName(WorkbookFileNameOrObj)
        Case "Workbook": Set m_ParentWorkbook = WorkbookFileNameOrObj
        Case "String"
            If Not IsWorkbookOpen(CStr(WorkbookFileNameOrObj), m_ParentWorkbook) Then
                On Error Resume Next
                Set m_ParentWorkbook = Workbooks.Open(Filename:=CStr(WorkbookFileNameOrObj), ReadOnly:=OpenReadOnly)
                On Error GoTo ErrorHandler
            End If
        Case Else: Set m_ParentWorkbook = ThisWorkbook
    End Select
    ValidateState checkWorkbook:=True

    ' Step 2: Set Worksheet
    Set m_ParentWorksheet = ResolveWorksheet(SheetNameOrObj)

    ' Step 3: Auto-detect ListObject if present
    If m_ParentWorksheet.ListObjects.Count > 0 Then
        If m_ParentWorksheet.ListObjects.Count > 1 Then
            RaiseError AWE_SheetError.ERR_InvalidTable, "Initialize", "Only one ListObject is supported per sheet."
        End If
        Set m_ListObject = m_ParentWorksheet.ListObjects(1)
        If Not IsValidListObject Then RaiseError AWE_SheetError.ERR_InvalidTable, "Initialize", "Invalid ListObject."
        m_HdrRowNb = m_ListObject.HeaderRowRange.row
    End If

    ' Step 4: Validate Header Row
    If m_ListObject Is Nothing Then
        If HeaderRowNumber < 1 Or HeaderRowNumber > m_ParentWorksheet.rows.Count Then _
            RaiseError AWE_SheetError.ERR_HeaderRowOutOfRange, "Initialize", "Header row out of range."
        m_HdrRowNb = HeaderRowNumber
    End If

    ' Step 5: Ensure AutoFilter dropdowns are present
    On Error Resume Next
    RangeX.AutoFilterMode = True
    On Error GoTo ErrorHandler

    ' Step 6: Map Columns
    If IsSheetEmpty Then
        RaiseError AWE_SheetError.ERR_SheetEmpty, "Initialize", "Sheet empty. Call MapColumnNumbers after creating headers."
    Else
        MapColumnNumbers
    End If

    ' Initialization successful
    Set Initialize = Me
    Exit Function

ErrorHandler:
    RaiseError Err.Number, "Initialize", Err.Description
    Set Initialize = Nothing
End Function

'---------------------------------------------------------------------------------------------------
' MapColumnNumbers - Maps column names to column numbers.
' Params:  None
' Returns: None - Updates the internal dictionary with column name-to-number mappings.
' Notes:   Clears existing mappings, validates metadata, and remaps them from the header row.
'          Raises an error if duplicate column names are detected.
'---------------------------------------------------------------------------------------------------
Public Sub MapColumnNumbers()
    On Error GoTo ErrorHandler
    ValidateState checkMetadata:=True
    m_ColumnHdrDict.RemoveAll

    Dim arr As Variant: arr = HeaderRowRangeX.Value
    Dim colIndex As Long

    For colIndex = LBound(arr, 2) To UBound(arr, 2)
        Dim colName As String: colName = Trim(arr(1, colIndex))
        If Len(colName) > 0 Then
            If Not m_ColumnHdrDict.Exists(colName) Then
                m_ColumnHdrDict.Add colName, colIndex
            Else
                RaiseError AWE_SheetError.ERR_MetadataInitializationError, "MapColumnNumbers", "Duplicate column: " & colName
            End If
        End If
    Next colIndex
    Exit Sub

ErrorHandler:
    RaiseError Err.Number, "MapColumnNumbers", "Error mapping columns: " & Err.Description
End Sub

'---------------------------------------------------------------------------------------------------
' IsWorkbookOpen - Checks if a workbook is already open in Excel.
' Params:  sFullFileName (String) - The full file name of the workbook to check.
'          wb (Workbook, Optional) - Output workbook object if the file is open.
' Returns: Boolean - True if the workbook is open, False otherwise.
'---------------------------------------------------------------------------------------------------
Public Function IsWorkbookOpen(sFullFileName As String, Optional ByRef wb As Workbook) As Boolean
    Dim sWbNm As String: sWbNm = LCase$(Mid$(sFullFileName, InStrRev(sFullFileName, "\") + 1))
    For Each wb In Application.Workbooks
        If LCase$(wb.Name) = sWbNm Then IsWorkbookOpen = True: Exit Function
    Next
    Set wb = Nothing
End Function

'###################################################################################################
' Metadata - Boundary and other informtaion about visible data on the sheet
'###################################################################################################
'---------------------------------------------------------------------------------------------------
' ParentWorkbook - Retrieves the associated workbook object with validation.
'---------------------------------------------------------------------------------------------------
Public Property Get ParentWorkbook() As Workbook
    ValidateState checkWorkbook:=True
    Set ParentWorkbook = m_ParentWorkbook
End Property

'---------------------------------------------------------------------------------------------------
' ParentWorksheet - Retrieves the associated worksheet object with validation.
'---------------------------------------------------------------------------------------------------
Public Property Get ParentWorksheet() As Worksheet
    ValidateState checkWorksheet:=True
    Set ParentWorksheet = m_ParentWorksheet
End Property

'---------------------------------------------------------------------------------------------------
' ParentTable - Retrieves the associated worksheet's table (ListObject) with validation.
'---------------------------------------------------------------------------------------------------
Public Property Get ParentTable() As ListObject
    ValidateState checkWorksheet:=True
    Set ParentTable = m_ListObject
End Property

'---------------------------------------------------------------------------------------------------
' HeaderRowNumber - Returns the header row number (Long) for the initialized worksheet.
'---------------------------------------------------------------------------------------------------
Public Property Get HeaderRowNumber() As Long
    ValidateState checkWorksheet:=True
    HeaderRowNumber = m_HdrRowNb
End Property

'---------------------------------------------------------------------------------------------------
' HeaderRowRangeX - Returns the range of the header row for the initialized worksheet.
'---------------------------------------------------------------------------------------------------
Public Property Get HeaderRowRangeX() As Range
    On Error GoTo ErrorHandler
    ValidateState checkWorksheet:=True

    Set HeaderRowRangeX = m_ParentWorksheet.Range(m_ParentWorksheet.Cells(m_HdrRowNb, 1), _
                                        m_ParentWorksheet.Cells(m_HdrRowNb, LastColumnNumber))
    Exit Property

ErrorHandler:
    RaiseError Err.Number, "HeaderRowRangeX", "Error retrieving header range: " & Err.Description
End Property

'---------------------------------------------------------------------------------------------------
' DataBodyRangeX - Returns the visible data range of the worksheet, excluding the header row.
' Params:  CellType (XlCellType, Optional) - Type of cells to include (Default: xlCellTypeVisible).
'          RaiseSheetEmptyError (Boolean, Optional) - Raises an error if the sheet is empty (Default: False).
' Returns: Range - The visible data range or Nothing if no data rows are visible.
'---------------------------------------------------------------------------------------------------
Public Property Get DataBodyRangeX(Optional CellType As XlCellType = xlCellTypeVisible, _
                                   Optional RaiseSheetEmptyError As Boolean = False) As Range
    On Error GoTo ErrorHandler
    Dim visibleRange As Range

    ' Retrieve the data range based on ListObject or standard range
    On Error Resume Next
    If IsValidListObject Then
        ' ListObject's DataBodyRange for visible rows
        Set visibleRange = m_ListObject.DataBodyRange.SpecialCells(CellType)
    Else
        ' Dynamically construct range starting after the header
        Set visibleRange = m_ParentWorksheet.Range(m_ParentWorksheet.Cells(m_HdrRowNb + 1, 1), _
                                                   m_ParentWorksheet.Cells(LastRowNumber, LastColumnNumber))
        If Not visibleRange Is Nothing And CellType = xlCellTypeVisible Then
            Set visibleRange = visibleRange.SpecialCells(CellType)
        End If
    End If
    On Error GoTo 0

    ' Handle empty range scenarios
    If visibleRange Is Nothing Then
        If RaiseSheetEmptyError Then
            RaiseError ERR_SheetEmpty, "DataBodyRangeX", "No visible rows available in the sheet."
        End If
    End If

    ' Return the visible range
    Set DataBodyRangeX = visibleRange
    Exit Property

ErrorHandler:
    RaiseError Err.Number, "DataBodyRangeX", "Error retrieving data body range: " & Err.Description
End Property

'---------------------------------------------------------------------------------------------------
' RangeX - Returns the combined range of the header row and data body range.
'---------------------------------------------------------------------------------------------------
Public Property Get RangeX() As Range
    ValidateState checkWorksheet:=True

    ' Combine HeaderRowRangeX and DataBodyRangeX using Resize
    Set RangeX = HeaderRowRangeX.Resize(rowSize:=rowCount + 1)
End Property

'---------------------------------------------------------------------------------------------------
' LastRowNumber - Returns the last row number (Long) excluding ghost rows with options for visibility.
'---------------------------------------------------------------------------------------------------
Public Property Get LastRowNumber() As Long
    On Error GoTo ErrorHandler
    Dim rowRange As Range, area As Range, iArea As Long, lastRow As Long, iRow As Long
    lastRow = 0

    ' Determine the base range
    Select Case True
        Case IsValidListObject
            ' Handle ListObject case
            lastRow = m_ListObject.DataBodyRange.rows(m_ListObject.DataBodyRange.rows.Count).row
            LastRowNumber = IIf(lastRow > m_HdrRowNb, lastRow, m_HdrRowNb)
            Exit Property
            
        Case m_ParentWorksheet.AutoFilterMode
            ' Use constants to exclude ghost rows
            On Error Resume Next
            Set rowRange = m_ParentWorksheet.AutoFilter.Range.SpecialCells(xlCellTypeVisible)
            On Error GoTo ErrorHandler

        Case Else
            Set rowRange = m_ParentWorksheet.Cells.SpecialCells(xlCellTypeConstants)
    End Select
   
    ' Exclude ghost rows, if needed
    If Not rowRange Is Nothing Then
        For iArea = rowRange.Areas.Count To 1 Step -1
            Set area = rowRange.Areas(iArea)
            For iRow = area.rows.Count To 1 Step -1
                ' Check if the row in the current area has non-empty cells
                If Application.CountA(area.rows(iRow).Value) > 0 Then
                    lastRow = area.rows(iRow).row
                    LastRowNumber = IIf(lastRow > m_HdrRowNb, lastRow, m_HdrRowNb)
                    Exit Property
                End If
            Next iRow
        Next iArea
    End If
    
    ' Ensure the last row is greater than or equal to the header row
    LastRowNumber = IIf(lastRow > m_HdrRowNb, lastRow, m_HdrRowNb)
    Exit Property

ErrorHandler:
    LastRowNumber = 0 ' Default to 0 in case of an error
    RaiseError Err.Number, "LastRowNumber", "Error retrieving last row number: " & Err.Description
End Property

'---------------------------------------------------------------------------------------------------
' LastColumnNumber - Returns the last populated column (Long) in the worksheet.
' Notes:   Returns 1 if the worksheet is empty.
'---------------------------------------------------------------------------------------------------
Public Property Get LastColumnNumber() As Long
    Select Case True
        Case IsSheetEmpty:      LastColumnNumber = 1
        Case IsValidListObject: LastColumnNumber = m_ListObject.DataBodyRange.columns.Count
        Case Else:              LastColumnNumber = m_ParentWorksheet.Cells(m_HdrRowNb, _
                                    m_ParentWorksheet.columns.Count).End(xlToLeft).Column
    End Select
End Property

'---------------------------------------------------------------------------------------------------
' RowCount - Returns the total count of visible rows in the DataBodyRange.
' Notes:   - Returns 0 if there are no visible rows or the range is empty.
'---------------------------------------------------------------------------------------------------
Public Property Get rowCount() As Long
    On Error Resume Next
    Dim visibleRange As Range
    Dim area As Range

    ' Use DataBodyRange to determine visible rows
    Set visibleRange = DataBodyRangeX

    ' Check if DataBodyRangeX is valid and contains visible rows
    If Not visibleRange Is Nothing Then
        If visibleRange.Areas.Count = 1 Then
            ' Contiguous range: directly return row count
            rowCount = visibleRange.rows.Count
        Else
            ' Non-contiguous range: sum row counts for all areas
            For Each area In visibleRange.Areas
                rowCount = rowCount + area.rows.Count
            Next area
        End If
    Else
        rowCount = 0 ' No rows exist or no rows are visible
    End If
    On Error GoTo 0
End Property

'---------------------------------------------------------------------------------------------------
' IsSheetEmpty - Returns True if the worksheet is empty; otherwise false.
'---------------------------------------------------------------------------------------------------
Public Property Get IsSheetEmpty() As Boolean
    On Error GoTo ErrorHandler
    
    ValidateState checkWorksheet:=True
    IsSheetEmpty = (Application.WorksheetFunction.CountA(m_ParentWorksheet.UsedRange) = 0)
    Exit Property

ErrorHandler:
    RaiseError Err.Number, "IsSheetEmpty", "Error while checking if sheet is empty: " & Err.Description
End Property

'---------------------------------------------------------------------------------------------------
' IsTable - Returns True if the sheet is based on a Table (ListObject); otherwise, false
'---------------------------------------------------------------------------------------------------
Public Property Get IsTable() As Boolean
    IsTable = Not m_ListObject Is Nothing
End Property

'###################################################################################################
' Column Functions - Retrieve a column number or range(s).
'###################################################################################################
'---------------------------------------------------------------------------------------------------
' GetColumnNumber - returns an array of column names (0 based index).
'---------------------------------------------------------------------------------------------------
Public Property Get columnNames() As Variant
    columnNames = m_ColumnHdrDict.keys()
End Property

'---------------------------------------------------------------------------------------------------
' GetColumnNumber - Resolves the column number from a String (header name), Long, or Range.
' Params:  col (Variant) - Column identifier (String, Long, or Range).
' Returns: Long - The column number.
' Notes:   Raises an error for invalid column types or missing column names.
'---------------------------------------------------------------------------------------------------
Public Function GetColumnNumber(col As Variant) As Long
    On Error GoTo ErrorHandler

    Select Case TypeName(col)
        Case "Long", "Integer"
            GetColumnNumber = col ' Return column index directly
        Case "String"
            ' Combine .Exists and retrieval for efficiency
            If m_ColumnHdrDict.Exists(col) Then
                GetColumnNumber = m_ColumnHdrDict(col)
            Else
                RaiseError AWE_SheetError.ERR_ColumnNotFound, "GetColumnNumber", "Column not found: " & col
            End If
        Case "Range"
            GetColumnNumber = col.Column ' Resolve column from a Range
        Case Else
            RaiseError AWE_SheetError.ERR_InvalidTable, "GetColumnNumber", _
                       "Invalid column type. Must be String, Long, or Range. Provided type: " & TypeName(col)
    End Select
    Exit Function

ErrorHandler:
    RaiseError Err.Number, "GetColumnNumber", "Error resolving column: " & Err.Description
End Function


'---------------------------------------------------------------------------------------------------
' ColumnsX - Combines one or more specified columns into a single range.
' Params:  ColIdxOrName (ParamArray) - List of column indexes or names that will be combined.
' Returns: Range - A single combined range for all specified columns.
' Notes:   Supports both contiguous and non-contiguous data areas.
'          Raises an error if any column name is invalid or cannot retrieve the range.
'---------------------------------------------------------------------------------------------------
Public Function ColumnsX(ParamArray ColIdxOrName() As Variant) As Range
    Dim col As Variant, colIndex As Long
    Dim combinedRange As Range, columnRange As Range

    On Error GoTo ErrorHandler

    ' Validate worksheet
    ValidateState checkWorksheet:=True

    ' Enforce constraint: No multiple columns if rows are filtered
    If IsFiltered And UBound(ColIdxOrName) >= 1 Then
        RaiseError AWE_SheetError.ERR_InvalidTable, "ColumnsX", _
                   "Cannot retrieve multiple columns when rows are filtered."
    End If

    ' Build combined range
    For Each col In ColIdxOrName
        colIndex = GetColumnNumber(col) ' Resolve column number
        If combinedRange Is Nothing Then
            Set combinedRange = Intersect(DataBodyRangeX, DataBodyRangeX.Parent.columns(colIndex))
        Else
            Set combinedRange = Union(combinedRange, Intersect(DataBodyRangeX, DataBodyRangeX.Parent.columns(colIndex)))
        End If
    Next col

    ' Validate combined range
    If combinedRange Is Nothing Then
        RaiseError AWE_SheetError.ERR_InvalidTable, "ColumnsX", "Could not retrieve valid column range."
    End If

    Set ColumnsX = combinedRange
    Exit Function

ErrorHandler:
    RaiseError Err.Number, "ColumnsX", "Error retrieving column range: " & Err.Description
    Set ColumnsX = Nothing
End Function

'###################################################################################################
' Row Functions - Filter, Search, Sort
'###################################################################################################
'---------------------------------------------------------------------------------------------------
' FilterX - Applies filters to a specified column in the worksheet or ListObject.
' Params:  headerName (String)            - The name of the column to filter by.
'          criteria1 (Variant)            - Primary filtering criterion or an array of criteria.
'          criteria2 (Variant, Optional)  - Secondary filtering criterion (default: vbNullString).
'          criteriaOperator (XlAutoFilterOperator, Optional) - Operator between criteria1 and criteria2
' Returns: Boolean - True if the filter is applied successfully, False otherwise.
' Example: aweSh.FilterX "Status", Array("Completed", "In Progress") ' Filter by multiple values
'          aweSh.FilterX "Date", ">" & #1/1/2021#, "<" & #1/15/2021# ' Filter by date range
'          aweSh.FilterX "Revenue", criteria1:=5, criteriaOperator:=xlTop10Items   ' Top 5 Revenues
'          aweSh.FilterX "Revenue", criteria1:=5, criteriaOperator:=xlTop10Percent ' Top 5% Revenues
'---------------------------------------------------------------------------------------------------
Public Function FilterX(headerName As String, criteria1 As Variant, _
                        Optional criteria2 As Variant = vbNullString, _
                        Optional criteriaOperator As XlAutoFilterOperator = xlAnd) As Boolean
    Dim colIndex As Long, filterRange As Range
    On Error GoTo ErrorHandler

    ' Validate worksheet state
    ValidateState checkWorksheet:=True
    If m_ParentWorksheet.ProtectContents Then _
        RaiseError AWE_SheetError.ERR_ProtectedWorksheet, "FilterX", "Cannot apply filters to a protected worksheet."
    If IsSheetEmpty Then _
        RaiseError AWE_SheetError.ERR_SheetEmpty, "FilterX", "Sheet is empty. Filtering cannot proceed."

    ' Determine the filter range using HeaderRowRangeX
    Set filterRange = HeaderRowRangeX

    ' Resolve the column index using GetColumnNumber
    colIndex = GetColumnNumber(headerName)

    ' Apply filter criteria
    Select Case True
        Case IsArray(criteria1)
            filterRange.AutoFilter Field:=colIndex, criteria1:=criteria1, Operator:=xlFilterValues
        Case IsNull(criteria2) Or criteria2 = vbNullString
            filterRange.AutoFilter Field:=colIndex, criteria1:=criteria1
        Case Else
            filterRange.AutoFilter Field:=colIndex, criteria1:=criteria1, _
                                   Operator:=criteriaOperator, criteria2:=criteria2
    End Select

    FilterX = True
    Exit Function

ErrorHandler:
    RaiseError Err.Number, "FilterX", "Error applying filter: " & Err.Description
    FilterX = False
End Function

'---------------------------------------------------------------------------------------------------
' ShowAllDataX - Removes all filters applied to the worksheet or worksheet ListObject.
' Params:  ShowErrMessage (Boolean, Optional) - Displays a message if the worksheet is protected
'                                           (default: True).
' Returns: Boolean - True if filters are cleared successfully or no filters exist, False otherwise.
' Notes:   Set ShowErrMessage to True for user feedback in interactive workflows.
'          Set ShowErrMessage to False to suppress messages in automated workflows.
'---------------------------------------------------------------------------------------------------
Public Function ShowAllDataX(Optional ShowErrMessage As Boolean = True) As Boolean
    On Error Resume Next

    If m_ParentWorksheet.ProtectContents Then
        If ShowErrMessage Then MsgBox "Sheet is protected. Unprotect to clear filters.", vbCritical, "AWE_Sheet"
    ElseIf IsValidListObject Then
        ' Clear filters for the ListObject
        If Not m_ListObject.AutoFilter Is Nothing Then m_ListObject.AutoFilter.ShowAllData
        ShowAllDataX = Err.Number = 0
    ElseIf Not m_ParentWorksheet.AutoFilter Is Nothing Then
        ' Clear filters for worksheet AutoFilter
        m_ParentWorksheet.ShowAllData
        ShowAllDataX = Err.Number = 0
    Else
        ' No filters to clear
        ShowAllDataX = True
    End If

    On Error GoTo 0
End Function

'---------------------------------------------------------------------------------------------------
' IsFiltered - Returns True if filtering is active on the worksheet or ListObject; otherwise, false.
'---------------------------------------------------------------------------------------------------
Public Property Get IsFiltered() As Boolean
    Dim filterRange As Range
    Dim col As Long

    On Error GoTo ErrorHandler
    IsFiltered = False
    ValidateState checkWorkbook:=True, checkWorksheet:=True

    ' Get the unified range
    Set filterRange = RangeX
    If filterRange Is Nothing Then Exit Property
    If m_ParentWorksheet.AutoFilter Is Nothing Then Exit Property

    ' Check for active filters
    For col = 1 To filterRange.columns.Count
        If m_ParentWorksheet.AutoFilter.Filters(col).On Then
            IsFiltered = True
            Exit Property
        End If
    Next col

    Exit Property

ErrorHandler:
    RaiseError Err.Number, "IsFiltered", "Error determining filter status: " & Err.Description
End Property

'---------------------------------------------------------------------------------------------------
' FindRowNumbersBase - Finds row numbers matching search values in the specified column.
' Params:  srchCol (Variant)    - Column to search in, specified by name or index.
'          FirstOnly (Boolean)  - If True, stops after finding the first match.
'          RaiseSearchError (Boolean) - If True, raises an error if no matches are found.
'          searchValues (ParamArray) - Values to search for in the column.
' Returns: Collection - Row numbers matching the search values.
' Notes:   - Searches all values unless FirstOnly = True.
'          - Raises ERR_SearchNotFound if no matches and RaiseSearchError = True.
'---------------------------------------------------------------------------------------------------
Private Function FindRowNumbersBase(srchCol As Variant, FirstOnly As Boolean, RaiseSearchError As Boolean, _
                                    ParamArray searchValues() As Variant) As Collection
    Dim combinedRange As Range, searchValue As Variant, result As New Collection
    Dim foundCell As Range, firstAddress As String

    ' Validate worksheet and protection
    ValidateState checkWorksheet:=True
    If m_ParentWorksheet.ProtectContents Then _
        RaiseError AWE_SheetError.ERR_ProtectedWorksheet, "FindRowNumbersBase", _
                   "Cannot perform lookups on a protected worksheet."
    If IsSheetEmpty Then _
        RaiseError AWE_SheetError.ERR_SheetEmpty, "FindRowNumbersBase", _
                   "Sheet is empty. Lookups cannot proceed."

    ' Early exit if no search values provided
    If UBound(searchValues) < 0 Then Exit Function ' No search values to process

    ' Retrieve combined column range
    Set combinedRange = ColumnsX(srchCol)

    ' Search in range for matching values
    For Each searchValue In searchValues
        If Not IsEmpty(searchValue) Then
            Set foundCell = combinedRange.Find(What:=searchValue, LookIn:=xlValues, LookAt:=xlPart)
            If Not foundCell Is Nothing Then
                firstAddress = foundCell.Address
                Do
                    result.Add foundCell.row
                    If FirstOnly Then Exit For ' Exit loop if FirstOnly is True
                    Set foundCell = combinedRange.FindNext(foundCell)
                Loop While Not foundCell Is Nothing And foundCell.Address <> firstAddress
            End If
        End If
    Next searchValue

    If result.Count = 0 And RaiseSearchError Then
        On Error Resume Next
        Dim joinVals As String: joinVals = Join(Application.WorksheetFunction.Transpose(searchValues), ", ")
        If Err.Number <> 0 Then joinVals = "Invalid input"
        On Error GoTo ErrorHandler
        RaiseError AWE_SheetError.ERR_SearchNotFound, "FindRowNumbersBase", _
                   "No matching data found in column: " & srchCol & " for values: " & joinVals
    End If


    ' Return results
    Set FindRowNumbersBase = result
    Exit Function

ErrorHandler:
    RaiseError Err.Number, "FindRowNumbersBase", "Error in FindRowNumbersBase: " & Err.Description
End Function

'---------------------------------------------------------------------------------------------------
' FindRowNumbers - Retrieves row numbers for matching search values in a specified column.
' Params:  srchCol (Variant) - The name or index of the column to search.
'          RaiseSearchError (Boolean) - Raise error if no matches are found.
'          searchValues (ParamArray) - List of values to search for.
' Returns: Collection - A collection of row numbers matching the search values.
' Example: aweSh.FindRowNumbers("Employee Name", True, "John*", "Mary*") ' Find rows for "John*" or "Mary*"
'---------------------------------------------------------------------------------------------------
Public Function FindRowNumbers(srchCol As Variant, RaiseSearchError As Boolean, _
                               ParamArray searchValues() As Variant) As Collection
    Set FindRowNumbers = FindRowNumbersBase(srchCol, False, RaiseSearchError, searchValues)
End Function

'---------------------------------------------------------------------------------------------------
' CellsX - Retrieves a range object based on row criteria and search/return columns.
' Params:  rowCriteria (Variant)       - Row identifier: a row number, range, or search value.
'          srchCol (Variant)           - Column to search in, specified by name or index.
'          rtrnCol (Variant, Optional) - Column to return (default: same as srchCol).
'          RaiseSearchError (Boolean, Optional) - Raise error if no match is found (default: True).
' Returns: Range - The resolved cell range, or Nothing if no match is found and errors are suppressed.
' Notes:   Raises ERR_SearchNotFound if no match is found and RaiseSearchError = True.
' Example: aweSh.CellsX(25, 1) ' Retrieve Row 25, Col 1
'          aweSh.CellsX("Mark Watson", "Employee Name") ' Find Mark Watson
'          aweSh.CellsX("Mary Wilson", "Employee Name", "Email") ' Lookup Email for Mary Wilson
'          aweSh.CellsX("Invalid Value", "Employee Name", , False) ' Suppress errors
'---------------------------------------------------------------------------------------------------
Public Function CellsX(rowCriteria As Variant, srchCol As Variant, _
                       Optional rtrnCol As Variant = vbNullString, _
                       Optional RaiseSearchError As Boolean = True) As Range
    Dim srchColNbr As Long, rtnColNbr As Long, rowNbr As Long
    Dim searchResults As Collection

    On Error GoTo ErrorHandler

    ' Resolve search and return column numbers
    srchColNbr = GetColumnNumber(srchCol)
    If rtrnCol = "" Then rtnColNbr = srchColNbr Else rtnColNbr = GetColumnNumber(rtrnCol)

    ' Determine row number from rowCriteria
    Select Case TypeName(rowCriteria)
        Case "Long", "Integer": rowNbr = rowCriteria ' Direct row number
        Case "Range": rowNbr = rowCriteria.row ' Row from range reference
        Case "String"
            ' Find row numbers using the base function (handles RaiseSearchError)
            Set searchResults = FindRowNumbersBase(srchColNbr, True, RaiseSearchError, rowCriteria)
            If searchResults.Count > 0 Then rowNbr = searchResults(1) Else rowNbr = 0
        Case Else
            ' Raise error for unsupported rowCriteria types
            RaiseError AWE_SheetError.ERR_InvalidParameter, "CellsX", _
                       "Unsupported rowCriteria type: " & TypeName(rowCriteria)
    End Select

    ' Validate and return the resolved cell
    If rowNbr > 0 And rowNbr <= m_ParentWorksheet.rows.Count Then
        Set CellsX = m_ParentWorksheet.Cells(rowNbr, rtnColNbr)
    ElseIf RaiseSearchError Then
        RaiseError AWE_SheetError.ERR_SearchNotFound, "CellsX", _
                   "Invalid row number: " & rowNbr & ". Row number must be between 1 and " & m_ParentWorksheet.rows.Count & "."
    Else
        Set CellsX = Nothing
    End If
    
    Exit Function

ErrorHandler:
    If Err.Number = AWE_SheetError.ERR_SearchNotFound Then
        RaiseError Err.Number, "CellsX", "Search String Not Found: '" & rowCriteria & "'"
    Else
        RaiseError Err.Number, "CellsX", "Error in CellsX: " & Err.Description
    End If
End Function

'---------------------------------------------------------------------------------------------------
' SortX - Sorts rows in the worksheet by one or more columns in ascending or descending order.
' Params:  ColIdxOrNames (ParamArray) - Column names or indices to sort by.
' Returns: Boolean - True if the sort is successful, False otherwise.
' Notes:   Example - SortX("Name", "<Age") sorts "Name" in ascending order and "Age" in descending order.
' Example: aweSh.SortX ">Week", "<Revenue" ' Multi-column sort - Ascending Week, Descending Revenue
'---------------------------------------------------------------------------------------------------
Public Function SortX(ParamArray ColIdxOrNames() As Variant) As Boolean
    Dim ColIdxOrName As Variant, colNumber As Long
    On Error GoTo ErrorHandler

    ' Validate worksheet state
    ValidateState checkWorksheet:=True
    If IsSheetEmpty Then Exit Function
    If m_ParentWorksheet.ProtectContents Then _
        RaiseError AWE_SheetError.ERR_ProtectedWorksheet, "SortX", "Sorting is not allowed on a protected worksheet."

    ' Configure sorting
    With m_ParentWorksheet.Sort
        For Each ColIdxOrName In ColIdxOrNames
            ' Resolve column number using GetColumnNumber
            colNumber = GetColumnNumber(Replace(Replace(ColIdxOrName, "<", ""), ">", ""))

            ' Add sort field
            .SortFields.Add key:=DataBodyRangeX.columns(colNumber), _
                            Order:=IIf(Left(ColIdxOrName, 1) = "<", xlDescending, xlAscending)
        Next
        .SetRange DataBodyRangeX
        .header = xlNo
        .Apply
        .SortFields.Clear
    End With

    SortX = True
    Exit Function

ErrorHandler:
    RaiseError Err.Number, "SortX", "Error during sorting: " & Err.Description
    SortX = False
End Function

'###################################################################################################
' Helper Functions - Modular private functions used in this class
'###################################################################################################
'---------------------------------------------------------------------------------------------------
' ResolveWorksheet - Resolves the specified worksheet by name, object, or table.
' Params:  vTarget (Variant) - Worksheet name, object, or ListObject to resolve.
' Returns: Worksheet - The resolved worksheet object.
' Notes:   Raises an error if the target cannot be resolved or is invalid.
'---------------------------------------------------------------------------------------------------
Private Function ResolveWorksheet(vTarget As Variant) As Worksheet
    On Error GoTo ErrorHandler

    If TypeName(vTarget) = "Worksheet" Then
        ' Directly assign if already a worksheet object
        Set ResolveWorksheet = vTarget
    ElseIf TypeName(vTarget) = "String" Then
        ' Resolve by name in the current workbook
        On Error Resume Next
        Set ResolveWorksheet = m_ParentWorkbook.Sheets(CStr(vTarget))
        On Error GoTo 0
    End If

    ' Validate resolved worksheet
    If ResolveWorksheet Is Nothing Then _
        RaiseError AWE_SheetError.ERR_InvalidWorksheet, "ResolveWorksheet", "Worksheet is not initialized."
        
    Exit Function

ErrorHandler:
    RaiseError Err.Number, "ResolveWorksheet", "Error retrieving worksheet: " & Err.Description
End Function

'---------------------------------------------------------------------------------------------------
' IsValidListObject - Returns True if the Table (ListObject) is valid; otherwise, false.
'---------------------------------------------------------------------------------------------------
Private Property Get IsValidListObject() As Boolean
    On Error Resume Next
    If Not m_ListObject Is Nothing Then
       If m_ListObject.ListColumns.Count > 0 Then IsValidListObject = True
    End If
    On Error GoTo 0
End Property

'---------------------------------------------------------------------------------------------------
' ValidateState - Consolidates validations for workbook, worksheet, and metadata states.
' Params:  checkWorkbook (Optional) - Validates that the workbook is initialized.
'          checkWorksheet (Optional) - Validates that the worksheet is initialized.
'          checkMetadata (Optional) - Validates that column header metadata exists and is populated.
'---------------------------------------------------------------------------------------------------
Private Sub ValidateState(Optional checkWorkbook As Boolean = False, _
                          Optional checkWorksheet As Boolean = False, _
                          Optional checkMetadata As Boolean = False)
    If checkWorkbook And m_ParentWorkbook Is Nothing Then _
        RaiseError AWE_SheetError.ERR_WorkbookNotFound, "ValidateState", "Workbook is not initialized."
    If checkWorksheet And m_ParentWorksheet Is Nothing Then _
        RaiseError AWE_SheetError.ERR_InvalidWorksheet, "ValidateState", "Worksheet is not initialized."
    If checkMetadata Then
        If HeaderRowRangeX Is Nothing Or Application.WorksheetFunction.CountA(HeaderRowRangeX) = 0 Then _
            RaiseError AWE_SheetError.ERR_MetadataInitializationError, "ValidateState", "Invalid or empty metadata."
    End If
End Sub




